/* crtecc: modulous error correction code test program v.0.2 Nov. 7, 2004
 * Copyright (c) 2004, Hosoda Takayuki. All rights reserved.
 *
 * This program demonstrates an error correction code utilizing the
 * chinese remainder theorem, which encodes 16 bit information data
 * to a set of five ASCII charactors. This code is capable to correct
 * single charactor error, and to detect two charactor errors.
 *
 * http://www.finetune.co.jp/~lyuka/ mailto:lyuka@finetune.co.jp
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#define NUM_VECT               5
#define NUM_SET                3
#define BIT_MASK      0x40000000L
#define DATA_LEN              16
#define DATA_MASK     0x0000ffffL
#define N2D           0x00010000L

long vect[NUM_VECT] = {38, 41, 43, 45, 47};
long vmax[NUM_VECT] = {66994, 79335, 90945, 80370, 73226};
long inv[NUM_VECT][NUM_SET] = {{58179, 55556, 20254}, {69660, 59040, 29971},
        {57105, 22231, 11610}, {74025, 28576, 58140}, {59737, 44650, 42066}};
int  vsel[NUM_VECT][NUM_SET] =
        {{0, 1, 2}, {1, 2, 3}, {2, 3, 4}, {0, 3, 4}, {0, 1, 4}};
char *cmap =
        "ABCDEFGHJKLMNPQRSTUVWXYZabcdefghjkmnpqrstuvwxyz0123456789@=+-/*#";
char tmap[128];

void
init()
{
    extern char     tmap[];
    extern char    *cmap;
    int             i;

    for (i = 0; i < 128; i++)
        tmap[i] = '?';
    for (i = 0; i < 64; i++)
        tmap[(int)cmap[i]] = i;
}

long
encode(d)
    long            d;
{
    long            code;
    extern long     vect[];

    d &= DATA_MASK;
    code = (((d % vect[0])) | ((d % vect[1]) << 6) | ((d % vect[2]) << 12)
        | ((d % vect[3]) << 18) | ((d % vect[4]) << 24));
    return code;
}

long
decode(code)
    long            code;
{
    int             i, j;
    long            d, data;
    long            a[NUM_VECT];
    long            b[NUM_VECT];

    for (i = 0; i < NUM_VECT; i++)
        a[i] = ((code >> (6 * i)) & 0x3f);
    for (i = 0; i < NUM_VECT; i++) {
        d = 0;
        for (j = 0; j < NUM_SET; j++)
            d = (d + (a[vsel[i][j]] * inv[i][j]) % vmax[i]) % vmax[i];
        b[i] = d;
    }
    if ((b[0] ^ b[1] ^ b[2] ^ b[4]) == 0) {
        data = b[0];
    } else {
        if (b[0] == b[1]) {
            data = b[0];
        } else if (b[2] == b[3]) {
            data = b[2];
        } else if (b[1] == b[2]) {
            data = b[1];
        } else if (b[3] == b[4]) {
            data = b[3];
        } else if (b[4] == b[0]) {
            data = b[4];
        } else {
            data = 0xffffffffL;
        }
        data |= 0x80000000L;
    }
    return data;
}

char *
bin2text(code, s)
    long            code;
    char           *s;
{
    int             i;
    extern char    *cmap;

    for (i = 0; i < NUM_VECT; i++)
        s[i] = cmap[(code >> (6 * i)) & 0x3f];
    s[NUM_VECT] = '\0';
    return s;
}

long
text2bin(text)
    char           *text;
{
    int             i;
    long            code;
    extern char     tmap[];

    code = 0;
    for (i = 0; i < NUM_VECT; i++)
        code |= ((long)tmap[(int)text[i]]) << (6 * i);
    return code;
}

#define MAX_TEST 10
int
main(argc, argv)
    int             argc;
    char           *argv[];
{
    int             i, ecount;
    long            data, code, ecode, cdata;
    char            txstr[NUM_VECT + 1];
    char            rxstr[NUM_VECT + 1];

    init();
    ecount = 0;
    for (i = 0; i < MAX_TEST; i++) {
        data = random() & DATA_MASK;
        code = encode(data);
        bin2text(code, txstr);
        strncpy(rxstr, txstr, NUM_VECT + 1);
        rxstr[(int)(random() % NUM_VECT)] ^= (random() & 0x3f);
        ecode = text2bin(rxstr);
        cdata = decode(ecode);

        printf("data = 0x%04lx, code = %s, ", data, txstr);
        printf("code+error = %s, ", bin2text(ecode, rxstr));
        if (cdata != 0xffffffffL) {
            if (cdata & 0x80000000L) {
                cdata &= ~0x80000000L;  /* clear error flag */
                printf("data = 0x%04lx(corrected)\n", cdata);
            } else {
                printf("data = 0x%04lx(no error)\n", cdata);
            }
        } else {
            printf("data = 0x%04lx(unrecoverble)\n", cdata);
            ecount++;
        }
    }
    printf("%d of %d SEC test passed.\n", MAX_TEST - ecount, MAX_TEST);
    return ecount;
}
